#  BSD 3-Clause License
#
#  Copyright (c) 2012, the Sentry Team, see AUTHORS for more details
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE

import base64
import os
import re
from functools import partial

from hoss_agent.conf import constants
from hoss_agent.utils import compat, encoding

try:
    from functools import partialmethod

    partial_types = (partial, partialmethod)
except ImportError:
    # Python 2
    partial_types = (partial,)


default_ports = {"https": 443, "http": 80, "postgresql": 5432}


def varmap(func, var, context=None, name=None):
    """
    Executes ``func(key_name, value)`` on all values,
    recursively discovering dict and list scoped
    values.
    """
    if context is None:
        context = set()
    objid = id(var)
    if objid in context:
        return func(name, "<...>")
    context.add(objid)
    if isinstance(var, dict):
        ret = func(name, dict((k, varmap(func, v, context, k)) for k, v in compat.iteritems(var)))
    elif isinstance(var, (list, tuple)):
        ret = func(name, [varmap(func, f, context, name) for f in var])
    else:
        ret = func(name, var)
    context.remove(objid)
    return ret


def is_master_process():
    # currently only recognizes uwsgi master process
    try:
        import uwsgi

        return os.getpid() == uwsgi.masterpid()
    except ImportError:
        return False


def get_url_dict(url):
    parse_result = compat.urlparse.urlparse(url)

    url_dict = {
        "full": encoding.keyword_field(url),
        "protocol": parse_result.scheme + ":",
        "hostname": encoding.keyword_field(parse_result.hostname),
        "pathname": encoding.keyword_field(parse_result.path),
    }

    port = None if parse_result.port is None else str(parse_result.port)

    if port:
        url_dict["port"] = port
    if parse_result.query:
        url_dict["search"] = encoding.keyword_field("?" + parse_result.query)
    return url_dict


def read_pem_file(file_obj):
    cert = b""
    for line in file_obj:
        if line.startswith(b"-----BEGIN CERTIFICATE-----"):
            break
    for line in file_obj:
        if not line.startswith(b"-----END CERTIFICATE-----"):
            cert += line.strip()
    return base64.b64decode(cert)



import platform


def get_runtime_env():
    return {
        "arch": platform.machine(),
        "hostname": platform.node() or "unknown",
        "platform:": platform.system(),
        "version": platform.python_version(),
    }




# helper to check if dictionary key is bytes and if so decode them into string, recursively
def bytes_key_to_string(d):
    rval = {}
    if not isinstance(d, dict):
        if isinstance(d, (tuple, list, set)):
            v = [bytes_key_to_string(x) for x in d]
            return v
        else:
            return d
    for k, v in d.items():
        if isinstance(k, bytes):
            k = k.decode()
        if isinstance(v, dict):
            v = bytes_key_to_string(v)
        elif isinstance(v, (tuple, list, set)):
            v = [bytes_key_to_string(x) for x in v]
        rval[k] = v
    return rval



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import decimal
import re
import base64


EVENTS_API_PATH = "v1"
AGENT_CONFIG_PATH = "/api/graphql"

TRACE_CONTEXT_VERSION = 0
TRACEPARENT_HEADER_NAME = "traceparent"
TRACESTATE_HEADER_NAME = "tracestate"

TIMESTAMP_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"

KEYWORD_MAX_LENGTH = 1024

HTTP_WITH_BODY = {"POST", "PUT", "PATCH", "DELETE"}
MAX_BODY_SIZE = 512000
MASK = 8 * "*"
ENCODED_MASK = base64.b64encode(MASK.encode('utf-8'))

EXCEPTION_CHAIN_MAX_DEPTH = 50

ERROR = "error"
TRANSACTION = "transaction"
SPAN = "span"
METRICSET = "metricset"
EVENT = "event"

LABEL_RE = re.compile('[.*"]')
HARDCODED_PROCESSORS = []

DEFAULT_SANITIZE_FIELD_NAMES = frozenset(
    ["authorization", "password", "secret", "passwd", "password", "token", "api_key", "access_token", "sessionid"]
)
HOST_BLACK_LIST = set(['app.hoss.com', 'app.hossapp.dev', "ingress.hoss.com", "ingress.hossapp.dev", 'apitracker.net', 'apitracker.com'])

try:
    # Python 2
    LABEL_TYPES = (bool, int, long, float, decimal.Decimal)
except NameError:
    # Python 3
    LABEL_TYPES = (bool, int, float, decimal.Decimal)

AGENT_CONFIG_GRAPHQL_QUERY = """
query AgentConfig {
  agentConfig {
    accountApiConfiguration {
      uuid
      hostBlacklist
      sanitizedHeaders
      sanitizedQueryParams
      sanitizedBodyFields {
        type
        value
      }
      bodyCapture
    }
    apis {
      uuid
      name
      rootDomain
      hosts
      logo
      configuration(mergeWithAccountConfiguration: true) {
        uuid
        sanitizedHeaders
        sanitizedQueryParams
        bodyCapture
        sanitizedBodyFields {
          type
          value
        }
      }
    }
  }
}
"""


import io
import time
from uuid import uuid4

import wrapt

from hoss_agent.conf.constants import EVENT, HOST_BLACK_LIST
from hoss_agent.context import execution_context
from hoss_agent.utils import compat
from hoss_agent.utils.logging import get_logger
from hoss_agent.utils.read_proxy import ReadProxy

try:
    from httplib import HTTPConnection, HTTPSConnection, HTTPResponse
except ImportError as error:
    from http.client import HTTPConnection, HTTPSConnection, HTTPResponse

logger = get_logger("hoss_agent.instrumentaion.packages.httpclient_hooks")


def install(module):
    wrapt.wrap_function_wrapper(
        module,
        "HTTPConnection._send_request", _sendrequest,
    )
    wrapt.wrap_function_wrapper(
        module,
        "HTTPConnection.getresponse", getresponse,
    )
    wrapt.wrap_function_wrapper(
        module,
        "HTTPResponse.close", response_close,
    )


# hook into sendrequest and set up a partial event object.
# the event object is store in the HTTPConnection object itself so it can be referenced in getresponse below
#
def _sendrequest(wrapped, instance, args, kwargs):
    #(self, method, url, body=None, headers={}, encode_chunked=False):
    try:
        host = instance.host
        method, url, body, headers = args[0:4]
        client = execution_context.get_client()
        if client is None:
            return wrapped(*args, **kwargs)
        if host in HOST_BLACK_LIST:
            logger.debug('Host is in blacklist. Skipping')
            return wrapped(*args, **kwargs)

        port = instance.port
        default_port = instance.default_port
        real_url = url

        if not real_url.startswith(("http://", "https://")):
            real_url = "%s://%s%s%s" % (
                default_port == 443 and "https" or "http",
                host,
                port != default_port and ":%s" % port or "",
                url,
            )
        request_received_at = int(round(time.time() * 1000))

        headers = dict(headers)

        req_body = io.BytesIO()
        if body is not None:
            if isinstance(body, bytes):
                req_body.write(body)
            elif isinstance(body, str):
                req_body.write(body.encode("utf-8"))
            else:
                # body is probably a file so we want to create a proxy around it to read its content
                largs = list(args)
                if hasattr(body, "read"):
                    proxy = ReadProxy(body, req_body)
                    largs[2] = proxy
                    args = tuple(largs)

        instance._hoss_context = {
            "event": {
                "eventId": uuid4(),
                "request": {
                    "method": method,
                    "headers": headers,
                    "body": req_body,
                    "url": real_url,
                    "receivedAt": request_received_at
                }
            },
            "client": client
        }
    except Exception as ex:
        logger.debug('Error caught in monkey-patched http client send_request: %s', ex)

    return wrapped(*args, **kwargs)


# Hook into getresponse to further construct the event object that was started in _sendrequest.
# We can't read response body here because it'll close the body. We need to hook into HTTPResponse.read for that.
# So we have to save the event to the response object so when read is called, we can reference the event
def getresponse(wrapped, instance, args, kwargs):
    rv = wrapped(*args, **kwargs)
    try:
        hoss_context = getattr(instance, "_hoss_context", None)

        if hoss_context is None:
            return wrapped(*args, **kwargs)

        response_received_at = int(round(time.time() * 1000))

        if compat.PY2:
            # in python 2, HTTPResponse headers is a list of raw http header output, so turn them into a dict by splitting by : then
            # trimming \r\n
            headers = {c[0]: c[1][0:-2] for c in [h.split(': ') for h in rv.msg.headers]}
        else:
            headers = dict(rv.headers)
        hoss_context['event']['response'] = {
            "headers": headers,
            "statusCode": rv.status,
            "receivedAt": response_received_at,
            "body": io.BytesIO()
        }
        rv._hoss_context = hoss_context
    except Exception as ex:
        logger.debug('Error in monkey-patched getresponse: %s', ex)

    return rv


# Monkey-patched response close. We will do final preparation steps for the event, e.g. base64 encode and queue it
# todo: Handle the case when close is never called. See the note below
#    # NOTE: it is possible that we will not ever call self.close(). This
#     #       case occurs when will_close is TRUE, length is None, and we
#     #       read up to the last byte, but NOT past it.
#     #
#     # IMPLIES: if will_close is FALSE, then self.close() will ALWAYS be
#     #          called, meaning self.isclosed() is meaningful.
def response_close(wrapped, instance, args, kwargs):
    rv = wrapped(*args, **kwargs)
    try:
        hoss_context = getattr(instance, "_hoss_context", None)
        if hoss_context is None:
            return rv

        if 'queued' not in hoss_context['event']:
            hoss_context['event']['queued'] = True
            hoss_context['client'].queue(EVENT, hoss_context['event'])
    except Exception as ex:
        logger.debug('Error in monkey-patched response.close: %s', ex)

    return rv



import wrapt


class BufferProxy(wrapt.ObjectProxy):
    __target__ = None

    def __init__(self, wrapped, target):
        super(BufferProxy, self).__init__(wrapped)
        self.__target__ = target

    def __iter__(self):
        it = iter(self.__wrapped__)
        for chunk in it:
            try:
                if isinstance(chunk, str):
                    self.__target__.write(chunk.encode("utf-8"))
                else:
                    self.__target__.write(chunk)
            except:
                pass
            yield chunk



import json
import unittest

import requests
from uuid import uuid4

from requests import ReadTimeout, ConnectionError

import hoss_agent
try:
    from unittest.mock import MagicMock, patch, ANY
except ImportError:
    from mock import MagicMock, patch, ANY

from hoss_agent.conf.constants import EVENT

try:
    from urllib.request import urlopen
except ImportError:
    from urllib import urlopen


# test httplib instrumentation by calling higher level http client method and verify that we can capture the
# call correctly
@patch('hoss_agent.transport.base.Transport.queue')
@patch('hoss_agent.transport.base.Transport.start_thread')
@patch('hoss_agent.Client.start_threads')
class TestRequests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        hoss_agent.init({
            "API_KEY": "API_KEY"
        })

    def test_basic_get(self, start_threads, transport_start_threads, queue):
        request_id = str(uuid4())
        url = "https://postman-echo.com/get?id=" + request_id
        response = requests.get(url)
        self.assertEqual(response.status_code, 200)
        queue.assert_called_once_with(EVENT, ANY, False)

        event_type = queue.call_args_list[0][0][0]
        event = queue.call_args_list[0][0][1]
        self.assertEqual(event_type, EVENT)

        self.assertEqual(event['request']['method'], 'GET')
        self.assertEqual(event['request']['url'], url)
        self.assertEqual(event['request']['body'].getvalue(), b'')

        self.assertEqual(event['response']['statusCode'], 200)
        self.assertEqual(request_id, response.json()['args']['id'])
        self.assertEqual(event['response']['headers'], dict(response.headers))

    def test_post_with_body(self, start_threads, transport_start_threads, queue):
        request_id = str(uuid4())
        url = "https://postman-echo.com/post?id=" + request_id
        response = requests.post(url, data=json.dumps({'key': 'value'}))
        self.assertEqual(response.status_code, 200)
        queue.assert_called_once_with(EVENT, ANY, False)

        event_type = queue.call_args_list[0][0][0]
        event = queue.call_args_list[0][0][1]
        self.assertEqual(event_type, EVENT)

        self.assertEqual(event['request']['method'], 'POST')
        self.assertEqual(event['request']['url'], url)
        self.assertEqual(event['request']['body'].getvalue(), b'{"key": "value"}')

        self.assertEqual(event['response']['statusCode'], 200)
        self.assertEqual(request_id, response.json()['args']['id'])
        self.assertEqual(event['response']['headers'], dict(response.headers))

    def test_chunked(self, start_threads, transport_start_threads, queue):
        url = "https://postman-echo.com/stream/1"
        response = requests.get(url)
        self.assertEqual(response.status_code, 200)
        queue.assert_called_once_with(EVENT, ANY, False)

        event_type = queue.call_args_list[0][0][0]
        event = queue.call_args_list[0][0][1]
        self.assertEqual(event_type, EVENT)

        self.assertEqual(event['response']['statusCode'], 200)
        body = json.loads(event['response']['body'].getvalue().decode())
        self.assertEqual(body['args']['n'], '1')

    def test_gzip(self, start_threads, transport_start_threads, queue):
        url = "https://postman-echo.com/gzip"
        response = requests.get(url)
        self.assertEqual(response.status_code, 200)
        queue.assert_called_once_with(EVENT, ANY, False)

        event_type = queue.call_args_list[0][0][0]
        event = queue.call_args_list[0][0][1]
        self.assertEqual(event_type, EVENT)

        self.assertEqual(event['response']['statusCode'], 200)
        body = json.loads(event['response']['body'].getvalue().decode())
        self.assertTrue(body['gzipped'])

    def test_deflate(self, start_threads, transport_start_threads, queue):
        url = "https://postman-echo.com/deflate"
        response = requests.get(url)
        self.assertEqual(response.status_code, 200)
        queue.assert_called_once_with(EVENT, ANY, False)

        event_type = queue.call_args_list[0][0][0]
        event = queue.call_args_list[0][0][1]
        self.assertEqual(event_type, EVENT)

        self.assertEqual(event['response']['statusCode'], 200)
        body = json.loads(event['response']['body'].getvalue().decode())
        self.assertTrue(body['deflated'])

    def test_timeout(self, start_threads, transport_start_threads, queue):
        url = "https://postman-echo.com/delay/2"
        try:
            requests.get(url, timeout=1)
        except ReadTimeout as ex:
            pass

        queue.assert_called_once_with(EVENT, ANY, False)
        event_type = queue.call_args_list[0][0][0]
        event = queue.call_args_list[0][0][1]
        self.assertEqual(event_type, EVENT)

        self.assertFalse('response' in event)
        self.assertEqual(event['error']['type'], 'CONNECTION_TIMEOUT')
        self.assertEqual(event['error']['context']["timeout"], 1000)

    def test_connection_error(self, start_threads, transport_start_threads, queue):
        url = "https://fasdftasdfa-echo.com/delay/2"
        try:
            response = requests.get(url, timeout=1)
        except ConnectionError as ex:
            pass

        queue.assert_called_once_with(EVENT, ANY, False)
        event_type = queue.call_args_list[0][0][0]
        event = queue.call_args_list[0][0][1]
        self.assertEqual(event_type, EVENT)

        self.assertFalse('response' in event)
        self.assertEqual(event['error']['type'], 'CONNECTION_ERROR')




#  BSD 3-Clause License
#
#  Copyright (c) 2012, the Sentry Team, see AUTHORS for more details
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE


import re

from hoss_agent.conf.constants import ERROR, MASK, SPAN, TRANSACTION, EVENT, ENCODED_MASK, MAX_BODY_SIZE
from hoss_agent.utils import compat, varmap
from hoss_agent.utils.encoding import force_text
import json

from hoss_agent.utils.logging import get_logger

logger = get_logger("hoss_agent.processors")

SANITIZE_FIELD_NAMES = frozenset(
    ["authorization", "password", "secret", "passwd", "password", "token", "api_key", "access_token", "sessionid"]
)

SANITIZE_VALUE_PATTERNS = [re.compile(r"^[- \d]{16,19}$")]  # credit card numbers, with or without spacers


def for_events(*events):
    """
    :param events: list of event types

    Only calls wrapped function if given event_type is in list of events
    """
    events = set(events)

    def wrap(func):
        func.event_types = events
        return func

    return wrap


def remove_body(client, api_config, event):
    """
    Removes body of request and response from context if body capture is off

    :param client: an hoss_agent client
    :param event: a transaction or error event
    :return: The modified event
    """
    try:
        if event is not None:
            body_capture = api_config[
                "body_capture"] if api_config is not None else client.client.config.account_api_configuration.get(
                'body_capture')
            if body_capture == "Off":
                logger.debug('Body capture is turned off. Removing body')
                event["request"]['body'] = ''
                if "response" in event and "body" in event["response"]:
                    event["response"]['body'] = ''
            elif body_capture == "OnError":
                not_has_error = "error" not in event
                has_successful_status_code = "response" in event and "status_code" in event["response"] and event["response"]["status_code"] < 400
                if not_has_error and has_successful_status_code:
                    logger.debug('Body capture is OnError. Removing body')
                    event["request"]['body'] = ''
                    event["response"]['body'] = ''

            if event["request"].get("body", None) is not None and len(event["request"]['body']) > MAX_BODY_SIZE:
                event["request"]['body'] = ''
                event["request"]['bodyLimitExceeded'] = True

            if 'response' in event and event["response"].get('body', None) is not None and len(event["response"]['body']) > MAX_BODY_SIZE:
                event["response"]['body'] = ''
                event["response"]['bodyLimitExceeded'] = True

        return event
    except Exception as ex:
        logger.debug('Error caught removing body %s', ex)
        return event

# def sanitize_http_request_cookies(client, api_config, event):
#     """
#     Sanitizes http request cookies
#
#     :param client: an hoss_agent client
#     :param event: a transaction or error event
#     :return: The modified event
#     """
#
#     # sanitize request.cookies dict
#     try:
#         cookies = event["request"]["cookies"]
#         event["request"]["cookies"] = varmap(_sanitize, cookies)
#     except (KeyError, TypeError):
#         pass
#
#     # sanitize request.header.cookie string
#     try:
#         cookie_string = event["request"]["headers"]["cookie"]
#         event["request"]["headers"]["cookie"] = _sanitize_string(cookie_string, "; ", "=")
#     except (KeyError, TypeError):
#         pass
#     return event


# def sanitize_http_response_cookies(client, api_config, event):
#     """
#     Sanitizes the set-cookie header of the response
#     :param client: an hoss_agent client
#     :param event: a transaction or error event
#     :return: The modified event
#     """
#     try:
#         cookie_string = event["response"]["headers"]["set-cookie"]
#         event["response"]["headers"]["set-cookie"] = _sanitize_string(cookie_string, ";", "=")
#     except (KeyError, TypeError):
#         pass
#     return event
#

def sanitize_http_headers(client, api_config, event):
    """
    Sanitizes http request/response headers

    :param client: an hoss_agent client
    :param event: a transaction or error event
    :return: The modified event
    """
    # request headers
    headers_to_sanitize = []
    try:
        headers_to_sanitize = api_config[
            "sanitized_headers"] if api_config is not None else client.client.config.account_api_configurations.get(
            'sanitized_headers')
    except:
        pass

    try:
        headers = event["request"]["headers"]
        event["request"]["headers"] = varmap(lambda k, v: _sanitize(k, v, headers_to_sanitize), headers)
    except (KeyError, TypeError):
        pass

    # response headers
    try:
        headers = event["response"]["headers"]
        event["response"]["headers"] = varmap(lambda k, v: _sanitize(k, v, headers_to_sanitize), headers)
    except (KeyError, TypeError):
        pass

    return event


def sanitize_http_request_querystring(client, api_config, event):
    """
    Sanitizes http request query string

    :param client: an hoss_agent client
    :param event: a transaction or error event
    :return: The modified event
    """
    try:
        url = force_text(event["request"]["url"], errors="replace")
    except (KeyError, TypeError):
        return event

    parsed_url = compat.urlparse.urlparse(url)
    query = parsed_url.query
    if "=" in query:
        params_to_sanitize = []
        try:
            params_to_sanitize = api_config[
                "sanitized_query_params"] if api_config is not None else client.client.config.account_api_configurations.get(
                'sanitized_query_params')
        except:
            pass

        sanitized_query_string = _sanitize_string(query, "&", "=", params_to_sanitize)
        event["request"]["url"] = url.replace(query, sanitized_query_string)
    return event


def _do_sanitize_http_body(body, fields):
    if fields and body:
        json_path_fields = map(lambda f: f['value'], filter(lambda f: f['type'] == 'JSONPath', fields))
        try:
            decoded_body = json.loads(body)
            return json.dumps(varmap(lambda k, v: _sanitize(k, v, json_path_fields), decoded_body)).encode('utf-8')
        except Exception as e:
            pass
    return body


def sanitize_http_body(client, api_config, event):
    """
    Sanitizes http request body. This only works if the request body
    is a query-encoded string. Other types (e.g. JSON) are not handled by
    this sanitizer.

    :param client: an hoss_agent client
    :param event: a transaction or error event
    :return: The modified event
    """
    try:
        request_body = force_text(event["request"]["body"], errors="replace")
        response_body = force_text(event["response"]["body"], errors="replace")
    except (KeyError, TypeError):
        return event
    fields_to_sanitize = []
    try:
        fields_to_sanitize = api_config[
            "sanitized_body_fields"] if api_config is not None else client.client.config.account_api_configurations.get(
            'sanitized_body_fields')
    except Exception as e:
        logger.debug('Error getting sanitize configuration: %s', e)
        return event
    try:
        event["request"]["body"] = _do_sanitize_http_body(request_body, fields_to_sanitize)
        event["response"]["body"] = _do_sanitize_http_body(response_body, fields_to_sanitize)
    except Exception as e:
        logger.debug('Error sanitizing http body: %s', e)
        return event
    return event


def _sanitize(key, value, fields=SANITIZE_FIELD_NAMES):
    if value is None:
        return

    if isinstance(value, compat.string_types) and any(pattern.match(value) for pattern in SANITIZE_VALUE_PATTERNS):
        return MASK

    if isinstance(value, dict):
        # varmap will call _sanitize on each k:v pair of the dict, so we don't
        # have to do anything with dicts here
        return value

    if not key:  # key can be a NoneType
        return value

    key = key.lower()
    for field in fields:
        if field.lower() in key:
            # store mask as a fixed length for security
            return MASK
    return value


def _sanitize_string(unsanitized, itemsep, kvsep, fields=SANITIZE_FIELD_NAMES):
    """
    sanitizes a string that contains multiple key/value items
    :param unsanitized: the unsanitized string
    :param itemsep: string that separates items
    :param kvsep: string that separates key from value
    :return: a sanitized string
    """
    sanitized = []
    kvs = unsanitized.split(itemsep)
    for kv in kvs:
        kv = kv.split(kvsep)
        if len(kv) == 2:
            sanitized.append((kv[0], _sanitize(kv[0], kv[1], fields)))
        else:
            sanitized.append(kv)
    return itemsep.join(kvsep.join(kv) for kv in sanitized)



#  BSD 3-Clause License
#
#  Copyright (c) 2012, the Sentry Team, see AUTHORS for more details
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE


from __future__ import absolute_import

import itertools
import os
import platform
import threading

import hoss_agent
from hoss_agent.conf import Config, VersionedConfig, constants
from hoss_agent.conf.constants import ERROR
from hoss_agent.utils import compat, is_master_process
from hoss_agent.utils.logging import get_logger
from hoss_agent.utils.module_import import import_string
from hoss_agent.context import execution_context

__all__ = ("Client",)


class Client(object):
    """
    The base hoss_agent client, which handles communication over the
    HTTP API to the ingress service.
    """

    logger = get_logger("hoss_agent")

    def __init__(self, config=None, **inline):
        # configure loggers first
        cls = self.__class__
        self.logger = get_logger("%s.%s" % (cls.__module__, cls.__name__))
        self.error_logger = get_logger("hoss_agent.errors")

        self._pid = None
        self._thread_starter_lock = threading.Lock()
        self._thread_managers = {}

        self.tracer = None
        self.processors = []
        self.filter_exception_types_dict = {}
        self._service_info = None

        config = Config(config, inline_dict=inline)
        if config.errors:
            for msg in config.errors.values():
                self.error_logger.error(msg)
            config.disable_send = True
        self.config = VersionedConfig(config, version=None)
        headers = {
            "Content-Type": "application/json",
            "Content-Encoding": "gzip",
            "User-Agent": "hoss_agent-python/%s" % hoss_agent.VERSION,
        }

        transport_kwargs = {
            "headers": headers,
            "verify_server_cert": self.config.verify_server_cert,
            "server_cert": self.config.server_cert,
            "timeout": self.config.server_timeout,
            "max_flush_time": self.config.api_request_time / 1000.0,
            "max_buffer_size": self.config.api_request_size,
            "processors": self.load_processors(),
        }
        self._api_endpoint_url = compat.urlparse.urljoin(
            self.config.server_url if self.config.server_url.endswith("/") else self.config.server_url + "/",
            constants.EVENTS_API_PATH,
        )
        self._config_endpoint_url = compat.urlparse.urljoin(
            self.config.config_server_url if self.config.config_server_url.endswith("/") else self.config.config_server_url + "/",
            constants.AGENT_CONFIG_PATH,
        )
        transport_class = import_string(self.config.transport_class)
        self._transport = transport_class(self._api_endpoint_url, self, **transport_kwargs)
        self.config.transport = transport_class(self._config_endpoint_url, self, **transport_kwargs)
        self._thread_managers["transport"] = self._transport

        if platform.python_implementation() == "PyPy":
            # PyPy introduces a `_functools.partial.__call__` frame due to our use
            # of `partial` in AbstractInstrumentedModule
            skip_modules = ("hoss_agent.", "_functools")
        else:
            skip_modules = ("hoss_agent.",)

        compat.atexit_register(self.close)
        if self.config.central_config:
            self._thread_managers["config"] = self.config
        else:
            self._config_updater = None

        execution_context.set_events([])
        execution_context.set_client(self)

    def start_threads(self):
        with self._thread_starter_lock:
            current_pid = os.getpid()
            if self._pid != current_pid:
                self.logger.debug("Detected PID change from %s to %d, starting threads", self._pid, current_pid)
                for manager_type, manager in self._thread_managers.items():
                    self.logger.debug("Starting %s thread", manager_type)
                    manager.start_thread()
                self._pid = current_pid

    def get_handler(self, name):
        return import_string(name)

    def capture(self, event_type, date=None, context=None, custom=None, stack=None, handled=True, **kwargs):
        """
        Captures and processes an event and pipes it off to Client.send.
        """
        if event_type == "Exception":
            # never gather log stack for exceptions
            stack = False
        data = self._build_msg_for_logging(
            event_type, date=date, context=context, custom=custom, stack=stack, handled=handled, **kwargs
        )

        if data:
            # queue data, and flush the queue if this is an unhandled exception
            self.queue(ERROR, data, flush=not handled)
            return data["id"]

    def capture_message(self, message=None, param_message=None, **kwargs):
        """
        Creates an event from ``message``.

        >>> client.capture_message('My event just happened!')
        """
        return self.capture("Message", message=message, param_message=param_message, **kwargs)

    def should_queue(self, data):
        for black_list_host in self.config.account_api_configuration['host_black_list']:
            if black_list_host in data['request']['url']:
                return False
        if self.config.disable_send:
            return False
        return True

    def flush(self):
        self._transport.flush()

    def queue(self, event_type, data, flush=False):
        if not self.should_queue(data):
            return
        self.start_threads()
        if flush and is_master_process():
            # don't flush in uWSGI master process to avoid ending up in an unpredictable threading state
            flush = False
        self._transport.queue(event_type, data, flush)

    def close(self):
        with self._thread_starter_lock:
            for _manager_type, manager in self._thread_managers.items():
                manager.stop_thread()

    def instrument(self):
        hoss_agent.instrument()

    def instrument(self):
        hoss_agent.instrument()

    def load_processors(self):
        """
        Loads processors from self.config.processors, as well as constants.HARDCODED_PROCESSORS.
        Duplicate processors (based on the path) will be discarded.

        :return: a list of callables
        """
        processors = itertools.chain(self.config.processors, constants.HARDCODED_PROCESSORS)
        seen = {}
        # setdefault has the nice property that it returns the value that it just set on the dict
        return [seen.setdefault(path, import_string(path)) for path in processors if path not in seen]


class DummyClient(Client):
    """Sends messages into an empty void"""

    def send(self, url, **kwargs):
        return None



import io
import wrapt


class ReadProxy(wrapt.ObjectProxy):
    __target__ = None

    def __init__(self, wrapped, target):
        super(ReadProxy, self).__init__(wrapped)
        self.__encode = isinstance(wrapped, io.TextIOBase)
        self.__target__ = target

    def read(self, amt=None):
        data = self.__wrapped__.read(amt)
        try:
            if self.__encode:
                self.__target__.write(data.encode("utf-8"))
            else:
                if isinstance(data, str):
                    self.__target__.write(data.encode("utf-8"))
                else:
                    self.__target__.write(data)
        except:
            pass
        return data



import base64
import time
from uuid import uuid4

import wrapt
from urllib3.exceptions import NewConnectionError, TimeoutError

from hoss_agent.conf.constants import EVENT
from hoss_agent.context import execution_context
from hoss_agent.utils.buffer_proxy import BufferProxy
from hoss_agent.utils.logging import get_logger

logger = get_logger("hoss_agent.instrumentaion.packages.urllib3_hooks")


def install(module):
    wrapt.wrap_function_wrapper(
        module,
        "connectionpool.HTTPConnectionPool._make_request", _make_request,
    )
    wrapt.wrap_function_wrapper(
        module,
        "response.HTTPResponse.read", read,
    )
    wrapt.wrap_function_wrapper(
        module,
        "response.HTTPResponse.read_chunked", read_chunked,
    )


# Instrument urllib3 connectionpool _make_request to capture connection error (both dns & connection timeout error)
# We're doing this at a higher level than httplib since the connection is httplib does't receive the full set of params
# (url, headers, etc...)
def _make_request(wrapped, instance, args, kwargs):
    request_received_at = int(round(time.time() * 1000))
    conn, method, url = args
    try:
        return wrapped(*args, **kwargs)
    except Exception as e:
        _capture_error(conn, method, instance.scheme, instance.host, url, kwargs.get('body'),
                       kwargs.get('headers'), request_received_at, e)
        raise e


def _capture_error(connection, method, scheme, host, url, body, headers, request_received_at, error):
    try:
        client = execution_context.get_client()
        if isinstance(error, NewConnectionError):
            error = {
                "type": 'CONNECTION_ERROR',
                "context": {
                    "error": str(error)
                },
                "receivedAt": int(round(time.time() * 1000))
            }
        elif isinstance(error, TimeoutError):
            error = {
                "type": 'CONNECTION_TIMEOUT',
                "context": {
                    "error": str(error),
                    "timeout": connection.timeout * 1000
                },
                "receivedAt": int(round(time.time() * 1000))
            }
        else:
            error = {
                "type": 'CONNECTION_ERROR',
                "context": {
                    "error": str(error),
                },
                "receivedAt": int(round(time.time() * 1000))
            }
        event = {
            "eventId": uuid4(),
            "request": {
                "method": method,
                "headers": dict(headers),
                "body": base64.b64encode(body) if body is not None else "",
                "url": "%s://%s%s" % (
                    scheme,
                    host,
                    url,
                ),
                "receivedAt": request_received_at
            },
            "error": error
        }
        client.queue(EVENT, event)
    except Exception as ex:
        logger.debug('Error capturing urllib.HTTPConnectionPool._make_request', ex)


# this hooks into urllib3.response
def read(wrapped, instance, args, kwargs):
    rv = wrapped(*args, **kwargs)
    try:
        hoss_context = getattr(instance._fp, "_hoss_context", None)
        if not hoss_context:
            return rv
        hoss_context['event']['response']['body'].write(rv)
        if instance.closed:
            logger.debug('Response is cloed after read. Queue event')
            if 'queued' not in hoss_context['event']:
                hoss_context['event']['queued'] = True
                hoss_context['client'].queue(EVENT, hoss_context['event'])
    except Exception as ex:
        logger.debug('Error in monkey-patched urllib3.response.read: %s', ex)

    return rv


# this is called to read an amt bytes out of the response. if hoss context doesn't have
# a response body yet, set it with the return value, otherwise concatenate
def read_chunked(wrapped, instance, args, kwargs):
    rv = wrapped(*args, **kwargs)
    try:
        hoss_context = getattr(instance._fp, "_hoss_context", None)
        if hoss_context is None:
            return rv

        rv = BufferProxy(rv,  hoss_context['event']['response']['body'])
    except Exception as ex:
        logger.debug('Error in monkey-patched urllib3.response.read_chunked: %s', ex)
    return rv




# -*- coding: utf-8 -*-

#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import base64
import hashlib
import re
import ssl

import certifi
import urllib3
from urllib3.exceptions import MaxRetryError, TimeoutError

from hoss_agent.conf import constants
from hoss_agent.transport.base import TransportException
from hoss_agent.transport.http_base import AsyncHTTPTransportBase, HTTPTransportBase
from hoss_agent.utils import compat, json_encoder, read_pem_file
from hoss_agent.utils.logging import get_logger

logger = get_logger("hoss_agent.transport.http")


class Transport(HTTPTransportBase):
    def __init__(self, url, *args, **kwargs):
        super(Transport, self).__init__(url, *args, **kwargs)
        url_parts = compat.urlparse.urlparse(url)
        pool_kwargs = {"cert_reqs": "CERT_REQUIRED", "ca_certs": certifi.where(), "block": True}
        if self._server_cert and url_parts.scheme != "http":
            pool_kwargs.update(
                {"assert_fingerprint": self.cert_fingerprint, "assert_hostname": False, "cert_reqs": ssl.CERT_NONE}
            )
            del pool_kwargs["ca_certs"]
        elif not self._verify_server_cert and url_parts.scheme != "http":
            pool_kwargs["cert_reqs"] = ssl.CERT_NONE
            pool_kwargs["assert_hostname"] = False
        proxies = compat.getproxies_environment()
        proxy_url = proxies.get("https", proxies.get("http", None))
        if proxy_url and not compat.proxy_bypass_environment(url_parts.netloc):
            self.http = urllib3.ProxyManager(proxy_url, **pool_kwargs)
        else:
            self.http = urllib3.PoolManager(**pool_kwargs)

    def send(self, data):
        response = None
        headers = self._headers.copy() if self._headers else {}
        headers.update({"Authorization": "Bearer " + self.client.config.api_key})
        if compat.PY2 and isinstance(self._url, compat.text_type):
            url = self._url.encode("utf-8")
        else:
            url = self._url
        try:
            try:
                response = self.http.urlopen(
                    "POST", url, body=data, headers=headers, timeout=self._timeout
                )
                logger.debug("Sent request, url=%s size=%.2fkb status=%s", url, len(data) / 1024.0, response.status)
            except Exception as e:
                print_trace = True
                if isinstance(e, MaxRetryError) and isinstance(e.reason, TimeoutError):
                    message = "Connection to Ingress Server timed out " "(url: %s, timeout: %s seconds)" % (
                        self._url,
                        self._timeout,
                    )
                    print_trace = False
                else:
                    message = "Unable to reach Ingress Server: %s (url: %s)" % (e, self._url)
                raise TransportException(message, data, print_trace=print_trace)
            body = response.read()
            if response.status >= 400:
                if response.status == 429:  # rate-limited
                    message = "Temporarily rate limited: "
                    print_trace = False
                else:
                    message = "HTTP %s: " % response.status
                    print_trace = True
                message += body.decode("utf8", errors="replace")
                raise TransportException(message, data, print_trace=print_trace)
            return response.getheader("Location")
        finally:
            if response:
                response.close()

    def get_config(self, current_version=None):
        """
        Gets configuration from a remote APM Server

        :param current_version: version of the current configuration
        :return: a three-tuple of new version, config dictionary and validity in seconds.
                 Any element of the tuple can be None.
        """
        url = self._url
        data = json_encoder.dumps({
            "query": constants.AGENT_CONFIG_GRAPHQL_QUERY
        }).encode("utf-8")
        headers = self._headers.copy()
        headers.update(self.auth_headers)
        max_age = 10
        try:
            response = self.http.urlopen(
                "POST", url, body=data, headers=headers, timeout=self._timeout, preload_content=False
            )
        except (urllib3.exceptions.RequestError, urllib3.exceptions.HTTPError) as e:
            logger.debug("HTTP error while fetching remote config: %s", compat.text_type(e))
            return current_version, None, max_age
        body = response.read()
        if "Cache-Control" in response.headers:
            try:
                max_age = int(next(re.finditer(r"max-age=(\d+)", response.headers["Cache-Control"])).groups()[0])
            except StopIteration:
                logger.debug("Could not parse Cache-Control header: %s", response.headers["Cache-Control"])
        if response.status >= 400:
            return None, None, max_age

        if not body:
            logger.debug("AgentConfig query returns empty body and status code %s", response.status)
            return current_version, None, max_age

        json_body = json_encoder.loads(body.decode("utf-8"))
        if 'errors' in json_body:
            logger.debug("AgentConfig query returns errors %s", json_body['errors'][0]['message'])
            return current_version, None, max_age
        # graphql api isn't returning 304 yet. so do the check manually
        etag = response.headers.get("Etag")
        if current_version is not None and current_version == etag:
            # config is unchanged, return
            logger.debug("Configuration unchanged")
            return current_version, None, max_age

        config = parse_remote_config(json_body['data']['agentConfig'])
        return etag, config, max_age

    @property
    def cert_fingerprint(self):
        if self._server_cert:
            with open(self._server_cert, "rb") as f:
                cert_data = read_pem_file(f)
            digest = hashlib.sha256()
            digest.update(cert_data)
            return digest.hexdigest()
        return None

    @property
    def auth_headers(self):
        headers = super(Transport, self).auth_headers
        return {k.encode("ascii"): v.encode("ascii") for k, v in compat.iteritems(headers)}


class AsyncTransport(AsyncHTTPTransportBase, Transport):
    async_mode = True
    sync_transport = Transport


def parse_api_configuration(api):
    return {
        "uuid": api["uuid"],
        "hosts": api["hosts"],
        "sanitized_headers": api["configuration"]["sanitizedHeaders"],
        "sanitized_query_params": api["configuration"]["sanitizedQueryParams"],
        "sanitized_body_fields": api["configuration"]["sanitizedBodyFields"],
        "body_capture": api["configuration"]["bodyCapture"]
    }


def parse_remote_config(remote_config):
    apis = map(parse_api_configuration, remote_config['apis'])
    api_host_mapping = {h: a for a in apis for h in a["hosts"]}
    return {
        "account_api_configuration": {
            "uuid": remote_config["accountApiConfiguration"]["uuid"],
            "host_black_list": remote_config["accountApiConfiguration"]["hostBlacklist"],
            "sanitized_headers": remote_config["accountApiConfiguration"]["sanitizedHeaders"],
            "sanitized_query_params": remote_config["accountApiConfiguration"]["sanitizedQueryParams"],
            "sanitized_body_fields": remote_config["accountApiConfiguration"]["sanitizedBodyFields"],
            "body_capture": remote_config["accountApiConfiguration"]["bodyCapture"]
        },
        "apis": apis,
        "api_host_mapping": api_host_mapping
    }



#  BSD 3-Clause License
#
#  Copyright (c) 2012, the Sentry Team, see AUTHORS for more details
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE


import os
import sys
import time
from threading import Lock, Thread

from hoss_agent.utils.compat import queue
from hoss_agent.utils.logging import get_logger

logger = get_logger("hoss_agent")

ELASTIC_APM_WAIT_SECONDS = 10


class AsyncWorker(object):
    _terminator = object()

    def __init__(self):
        self._queue = queue.Queue(-1)
        self._lock = Lock()
        self._thread = None
        self.start()

    def main_thread_terminated(self):
        self._lock.acquire()
        try:
            if not self._thread:
                # thread not started or already stopped - nothing to do
                return

            # wake the processing thread up
            self._queue.put_nowait(self._terminator)

            # wait briefly, initially
            initial_timeout = 0.1

            if not self._timed_queue_join(initial_timeout):
                # if that didn't work, wait a bit longer
                # NB that size is an approximation, because other threads may
                # add or remove items
                size = self._queue.qsize()

                sys.stdout.write("PID %i: hoss_agent is attempting to send %i pending messages\n" % (os.getpid(), size))
                sys.stdout.write("Waiting up to %s seconds, " % ELASTIC_APM_WAIT_SECONDS)
                wait_start = time.time()
                if os.name == "nt":
                    sys.stdout.write("press Ctrl-Break to quit.\n")
                else:
                    sys.stdout.write("press Ctrl-C to quit.\n")
                self._timed_queue_join(ELASTIC_APM_WAIT_SECONDS - initial_timeout)
                sys.stdout.write(
                    "PID %i: done, took %.2f seconds to complete.\n" % (os.getpid(), time.time() - wait_start)
                )
            self._thread = None

        finally:
            self._lock.release()

    def _timed_queue_join(self, timeout):
        """
        implementation of Queue.join which takes a 'timeout' argument

        returns True on success, False on timeout
        """
        deadline = time.time() + timeout

        with self._queue.all_tasks_done:
            while self._queue.unfinished_tasks:
                delay = deadline - time.time()
                if delay <= 0:
                    # timed out
                    return False

                self._queue.all_tasks_done.wait(timeout=delay)

            return True

    def start(self):
        """
        Starts the task thread.
        """
        self._lock.acquire()
        try:
            if not self._thread:
                self._thread = Thread(target=self._target)
                self._thread.setDaemon(True)
                self._thread.name = "hoss_agent sender thread"
                self._thread.start()
        finally:
            self._lock.release()

    def stop(self, timeout=None):
        """
        Stops the task thread. Synchronous!
        """
        self._lock.acquire()
        try:
            if self._thread:
                self._queue.put_nowait(self._terminator)
                self._thread.join(timeout=timeout)
                self._thread = None
        finally:
            self._lock.release()

    def is_alive(self):
        return self._thread is not None and self._thread.is_alive()

    def queue(self, callback, kwargs):
        self._queue.put_nowait((callback, kwargs))

    def _target(self):
        while True:
            try:
                record = self._queue.get()
                if record is self._terminator:
                    break
                callback, kwargs = record
                try:
                    callback(**kwargs)
                except Exception:
                    logger.error("Error while sending", exc_info=True)
            finally:
                self._queue.task_done()
            time.sleep(0)



#  BSD 3-Clause License
#
#  Copyright (c) 2012, the Sentry Team, see AUTHORS for more details
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
import sys

__all__ = ("VERSION", "Client")

try:
    VERSION = __import__("pkg_resources").get_distribution("hoss").version
except Exception:
    VERSION = "unknown"

from hoss_agent.base import Client
from hoss_agent.instrumentation.control import instrument
from hoss_agent.utils.logging import get_logger

GLOBAL_CLIENT = None
logger = get_logger("hoss_agent")


def init(config=None, *args, **inline):

    if isinstance(config, str):
        if len(config) == 0:
            logger.info('Hoss API Key is not provided. Exiting')
            return
        if len(args) >= 1 and isinstance(args[0], dict):
            args[0]['API_KEY'] = config
            client = Client(args[0], **inline)
        else:
            client = Client({"API_KEY": config}, **inline)
    else:
        if not config or not config['API_KEY']:
            logger.info('Hoss API Key is not provided. Exiting')
            return
        client = Client(config, **inline)

    client.instrument()
    client.start_threads()
    global GLOBAL_CLIENT
    GLOBAL_CLIENT = client
    return client
